#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''crucible

Submits diffs to Crucible.
'''
import os
import tempfile
from urlparse import urljoin
import xml.etree.ElementTree as ElementTree
from xml.sax.saxutils import escape
from mercurial import cmdutil, patch, util
from mercurial.i18n import _
import cStringIO
from restkit import request, BasicAuth
from helpers import get_between, memmerge, setworkingctx

CREATE_REVIEW_URL = 'rest-service/reviews-v1'
ADD_PATCH_URL = 'rest-service/reviews-v1/%s/patch'
ADD_REVIEWERS_URL = 'rest-service/reviews-v1/%s/reviewers'
APPROVE_URL = 'rest-service/reviews-v1/%s/transition?action=action:approveReview'

CREATE_REVIEW_XML_TEMPLATE = \
"""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<createReview>
    <reviewData>
        <allowReviewersToJoin>true</allowReviewersToJoin>
        <author>
            <userName>%s</userName>
        </author>
        <creator>
            <userName>%s</userName>
        </creator>
        <description>Description or statement of objectives for this example review.</description>
        <moderator>
            <userName>%s</userName>
        </moderator>
        <name>%s</name>
        <projectKey>%s</projectKey>
        <state>Review</state>
        <type>REVIEW</type>
    </reviewData>
</createReview>
"""

ADD_PATCH_XML_TEMPLATE = \
'''<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<addPatch>
    <patch>%s</patch>
</addPatch>
'''

class ReviewException(Exception):
    '''
    Raised when an error occurs during review creation.
    '''

    def __init__(self, message, response=None):
        self.message = message
        self.response = response

def _successful(response):
    return (response.status_int // 100) is 2

def review(ui, repo, *args, **opts):
    """
    Collects all applicable changesets and creates a review in Crucible.
    """
    rev = opts.get('rev')

    def create_review(title):
        """
        Hits Crucible's /reset-service/reviews-v1 endpoint to create a new,
        empty review.
        Returns the new review's permaId string.
        """
        url = urljoin(opts['host'], CREATE_REVIEW_URL)
        user = opts['username']
        moderator = opts['moderator'] if opts['moderator'] else user

        body = CREATE_REVIEW_XML_TEMPLATE % \
               (escape(user), escape(user), escape(moderator), escape(title[:255]), escape(opts['project']))

        resp = request(url,
                       method='POST',
                       body=body,
                       filters=[BasicAuth(user, opts['password'])],
                       headers=
                        {'Content-Type': 'application/xml',
                         'Accept': 'application/xml'},
                       follow_redirect=True)

        if _successful(resp):
            xml = resp.body_string()
            id = ElementTree.XML(xml).findtext('.//permaId/id')
            ui.write(_(urljoin(opts['host'], 'cru/' + id) + '\n'))
            ui.flush()
            return id
        else:
            raise ReviewException('Unable to create new review', resp)

    def add_patch(permaid, patch):
        """
        POSTs to Crucible's /rest-service/reviews-v1/{id}/patch endpoint to
        add one changeset to the review as a patch.
        """
        url = urljoin(opts['host'], (ADD_PATCH_URL % permaid))

        body = ADD_PATCH_XML_TEMPLATE % escape(patch)
        resp = request(url,
                       method='POST',
                       body=body,
                       filters=[BasicAuth(opts['username'], opts['password'])],
                       headers=
                        {'Content-Type': 'application/xml',
                         'Accept': 'application/xml'},
                       follow_redirect=True)

        if not _successful(resp):
            raise ReviewException('Unable to add changeset to review', resp)

    def add_reviewers(permaid):
        url = urljoin(opts['host'], (ADD_REVIEWERS_URL % permaid))

        resp = request(url,
                       method='POST',
                       body=opts['reviewers'],
                       filters=[BasicAuth(opts['username'], opts['password'])],
                       headers=
                        {'Content-Type': 'application/xml',
                         'Accept': 'application/xml'},
                       follow_redirect=True)

        if not _successful(resp):
            raise ReviewException('Unable to add reviewers (%s) to review' % opts['reviewers'], resp)

    def open_review(permaid):
        '''
        Takes the specified review from Draft to Under Review state.
        '''
        url = urljoin(opts['host'], APPROVE_URL % permaid)
        ui.write(_('Taking review to Open state at: ' + url + '...\n'))
        ui.flush()

        resp = request(url,
                       method='POST',
                       filters=[BasicAuth(opts['username'], opts['password'])],
                       headers=
                        {'Content-Type': 'application/xml',
                         'Accept': 'application/xml'},
                       follow_redirect=True)

        raise ReviewException('Unable to open review', resp)

    revs = cmdutil.revrange(repo, [rev])
    p = cStringIO.StringIO()
    if len(revs) == 1:
        cmdutil.export(repo, [revs[0]], fp=p,
                       opts=patch.diffopts(ui, opts))

    else:
        left_repo = repo
        right_repo = repo

        left_cs = left_repo[revs[0]]
        right_cs = right_repo[revs[-1]]

        bfd, bfn = tempfile.mkstemp(suffix='.hg',
                                    prefix='hg-bundle-')

        try:
            other, between, firstparent, last = get_between(left_repo, left_cs, right_repo, right_cs, bfn)
            if not between:
                raise util.Abort(_('no changes found'))

            try:
                ctx, status, merged, conflicts = memmerge(
                    other, right_cs.node(), left_cs.node(),
                    local_marker=u'LOCAL',
                    remote_marker=u'REMOTE',
                    limit=150)
                merge_limited = False
            except OverflowError:
                ctx, status, merged, conflicts = None, None, set(), set()
                merge_limited = True

            if ctx is not None or merge_limited:
                firstparent, last = right_cs, ctx

            # Do the diff
            if not merge_limited:
                with setworkingctx(other, ctx):
                    for chunk in patch.diff(other, firstparent.node(),
                                            last.node(),
                                            changes=status):
                        p.write(chunk)

        finally:
            os.unlink(bfn)


    def _get_config_value(opts, key, description, default=None, allowEmpty=False, echo=True):
        if not opts[key]:
            opts[key] = ui.config('review', key, default=None)
            while not opts[key]:
                opts[key] = ui.prompt(description, default) if echo else ui.getpass(description, default)
                if opts[key] or allowEmpty:
                    break

        return opts[key]

    opts['host'] = _get_config_value(opts, 'host', _('Crucible base url:'))
    if not opts['host'].endswith('/'):
        opts['host'] += '/'
    opts['username'] = _get_config_value(opts, 'username', _('username:'))
    opts['password'] = _get_config_value(opts, 'password', _('password: '), allowEmpty=True, echo=False)
    opts['reviewers'] = _get_config_value(opts, 'reviewers', _('reviewers (comma separated, e.g. "fred,joe,matt"): '), allowEmpty=True)
    opts['project'] = _get_config_value(opts, 'project', _('project (default is "CR"):'), default='CR')
    opts['unified'] = _get_config_value(opts, 'context', _('context lines (default is 100000):'), default=100000)

    try:
        id = create_review(repo[revs[0]].description())
        add_patch(id, p.getvalue())
        add_reviewers(id)
#        open_review(id)

    except ReviewException, re:
        if opts['vv']:
            ui.warn(re.response.body_string() + '\n')
        print re.message + ': ' + re.response.status
        if not opts['vv']:
            ui.warn('use --vv for full server responses on errors\n')
